#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Running verification workchain
"""
import os

import aiida
import click
from aiida import orm
from aiida.cmdline.params import options, types
from aiida.cmdline.utils import echo
from aiida.engine import run_get_node, submit
from aiida.plugins import DataFactory, WorkflowFactory

from aiida_sssp_workflow.cli import cmd_root
from aiida_sssp_workflow.workflows.verifications import (
    DEFAULT_CONVERGENCE_PROPERTIES_LIST,
    DEFAULT_MEASURE_PROPERTIES_LIST,
    DEFAULT_PROPERTIES_LIST,
)

UpfData = DataFactory("pseudo.upf")
VerificationWorkChain = WorkflowFactory("sssp_workflow.verification")

# Trigger the launch by running:
# aiida-sssp-workflow launch --property measure.precision --pw-code pw-7.0@localhost --ph-code ph-7.0@localhost --protocol test --cutoff-control test --criteria efficiency --withmpi True -- examples/_static/Si_ONCV_PBE-1.2.upf


@cmd_root.command("launch")
@click.argument("pseudo", type=click.Path(exists=True))
@options.OverridableOption(
    "--pw-code", "pw_code", type=types.CodeParamType(entry_point="quantumespresso.pw")
)(required=True)
@options.OverridableOption(
    "--ph-code", "ph_code", type=types.CodeParamType(entry_point="quantumespresso.ph")
)(required=False)
@click.option(
    "--property",
    multiple=True,
    default=[],
    help="Property to verify, can be: measure.precision, measure.bands, convergence ...",
)
@click.option(
    "protocol",
    "--protocol",
    default="acwf",
    help="Protocol to use for the verification, (acwf, test).",
)
@click.option("withmpi", "--withmpi", default=True, help="Run with mpi.")
@click.option("npool", "--npool", default=1, help="Number of pool.")
@click.option("walltime", "--walltime", default=3600, help="Walltime.")
@click.option("num_mpiprocs", "--num-mpiprocs", default=1, help="Number of mpiprocs.")
@click.option(
    "--clean-workdir/--no-clean-workdir",
    default=True,
    help="Clean up the remote folder of all calculation, turn this off when your want to see the remote for details.",
)
@click.option(
    "--daemon/--no-daemon",
    default=True,
    help="Launch the verfication to daemon by submit or run directly.",
)
@click.option(
    "comment",
    "--comment",
    help="Add a comment word as the prefix the workchain description.",
)
# ecutwfc and ecutrho are for measure workflows only
@click.option(
    "ecutwfc",
    "--ecutwfc",
    help="Cutoff energy for wavefunctions in Rydberg.",
)
@click.option(
    "ecutrho",
    "--ecutrho",
    help="Cutoff energy for charge density in Rydberg.",
)
# cutoff_control, criteria, configuration is for convergence workflows only
@click.option(
    "cutoff_control",
    "--cutoff-control",
    help="Cutoff control for convergence workflow, (standard, quick, opsp).",
)
@click.option(
    "criteria", "--criteria", help="Criteria for convergence (efficiency, precision)."
)
# configuration is hard coded for convergence workflow, but here is an interface for experiment purpose
@click.option(
    "configuration",
    "--configuration",
    help="(convergence test only) Configuration of structure, can be: SC, FCC, BCC, Diamond, XO, XO2, XO3, X2O, X2O3, X2O5, GS, RE",
)
def launch(
    pw_code,
    ph_code,
    property,
    protocol,
    ecutwfc,
    ecutrho,
    cutoff_control,
    criteria,
    configuration,
    withmpi,
    npool,
    walltime,
    num_mpiprocs,
    pseudo,
    clean_workdir,
    daemon,
    comment,
):
    """Launch the verification workchain."""
    # if the property is not specified, use the default list with all properties calculated.
    # otherwise, use the specified properties.
    if not property:
        properties_list = DEFAULT_PROPERTIES_LIST
        extra_desc = "All properties"
    elif len(property) == 1 and property[0] == "convergence":
        properties_list = DEFAULT_CONVERGENCE_PROPERTIES_LIST
        extra_desc = "Convergence"
    elif len(property) == 1 and property[0] == "measure":
        properties_list = DEFAULT_MEASURE_PROPERTIES_LIST
        extra_desc = "Measure"
    else:
        properties_list = list(property)
        extra_desc = f"{properties_list}"

    # validate the options are all provide for the property
    is_convergence = False
    is_measure = False
    is_ph = False
    for prop in properties_list:
        if prop.startswith("convergence"):
            is_convergence = True
        if prop.startswith("measure"):
            is_measure = True
        if "phonon_frequencies" in prop:
            is_ph = True

    # raise error if the options are not provided
    if is_convergence and not (cutoff_control and criteria):
        echo.echo_critical(
            "cutoff_control, criteria must be provided for convergence workflow."
        )

    if is_measure and not (ecutwfc and ecutrho):
        echo.echo_critical("ecutwfc and ecutrho must be provided for measure workflow.")

    if is_ph and not ph_code:
        echo.echo_critical("ph_code must be provided for phonon frequencies.")

    # raise warning if the options are over provided, e.g. cutoff_control is provided for measure workflow
    if is_measure and (cutoff_control or criteria or configuration):
        echo.echo_warning(
            "cutoff_control, criteria, configuration are not used for measure workflow."
        )

    if is_convergence and (ecutwfc or ecutrho):
        echo.echo_warning("ecutwfc and ecutrho are not used for convergence workflow.")

    _profile = aiida.load_profile()
    echo.echo_info(f"Current profile: {_profile.name}")

    basename = os.path.basename(pseudo)

    computer = pw_code.computer.label
    label, _ = os.path.splitext(basename)
    conf_label = configuration or "default"
    pre_label = (
        f"{protocol}"
        if not is_convergence
        else f"{protocol}-{criteria}-{cutoff_control}"
    )
    label = orm.Str(f"({pre_label} at {computer} - {conf_label}) {label}")

    with open(pseudo, "rb") as stream:
        pseudo = UpfData(stream)

    inputs = {
        "measure": {
            "protocol": orm.Str(protocol),
            "wavefunction_cutoff": orm.Float(ecutwfc),
            "charge_density_cutoff": orm.Float(ecutrho),
        },
        "convergence": {
            "protocol": orm.Str(protocol),
            "cutoff_control": orm.Str(cutoff_control),
            "criteria": orm.Str(criteria),
        },
        "pw_code": pw_code,
        "pseudo": pseudo,
        "label": label,
        "properties_list": orm.List(properties_list),
        "options": orm.Dict(
            dict={
                "resources": {
                    "num_machines": 1,
                    "num_mpiprocs_per_machine": num_mpiprocs,
                },
                "max_wallclock_seconds": walltime,
                "withmpi": withmpi,
            }
        ),
        "parallelization": orm.Dict(dict={"npool": npool}),
        "clean_workdir": orm.Bool(clean_workdir),
    }

    if is_ph:
        inputs["ph_code"] = ph_code

    if configuration is not None:
        inputs["convergence"]["configuration"] = orm.Str(configuration)
        inputs["measure"]["configurations"] = orm.List(list=[configuration])

    if daemon:
        node = submit(VerificationWorkChain, **inputs)
    else:
        _, node = run_get_node(VerificationWorkChain, **inputs)

    description = f"{label.value} ({extra_desc})"
    node.description = f"({comment}) {description}" if comment else description

    echo.echo_info(node)
    echo.echo_success(f"calculated on property: {'/'.join(properties_list)}")


if __name__ == "__main__":
    launch()



# -*- coding: utf-8 -*-
# pylint: disable=wrong-import-position,wildcard-import
"""Module for the command line interface."""
import click
from aiida.cmdline.groups import VerdiCommandGroup
from aiida.cmdline.params import options, types


@click.group(
    "aiida-sssp-workflow",
    cls=VerdiCommandGroup,
    context_settings={"help_option_names": ["-h", "--help"]},
)
@options.PROFILE(type=types.ProfileParamType(load_profile=True), expose_value=False)
def cmd_root():
    """CLI for the `aiida-sssp-workflow` plugin."""


from .extract import extract
from .run import launch



import click
from aiida.orm import load_node

from aiida_sssp_workflow.cli import cmd_root


@cmd_root.command("relabel")
@click.argument("label", type=str, nargs=1)
@click.argument("node", type=str, nargs=1)
def relabel_node(label, node):
    """Relabel verification node to extra"""
    _node = load_node(node)
    old_label = _node.extras.get("label")

    if label == old_label:
        click.echo("Please input a new label.")
    else:
        _node.set_extra("label", label)
        click.echo(
            f"uuid={_node.uuid} reset extra label {old_label} to {_node.extras.get('label')}"
        )



import pytest

from aiida_sssp_workflow.utils import get_standard_structure


def test_get_standard_structure_default(data_regression):
    """Test get_standard_structure function."""
    with pytest.raises(ValueError):
        structure = get_standard_structure("Si", "delta")

    structure = get_standard_structure("Si", "delta", configuration="fcc")
    data_regression.check(structure.get_cif().get_content())


@pytest.mark.parametrize(
    "element, prop, configuration",
    [
        ("Fe", "delta", "gs"),
        ("Fe", "convergence", None),
        ("Te", "convergence", "xo"),
        ("Te", "delta", "xo"),
    ],
)
def test_get_standard_structure(element, prop, configuration, data_regression):
    """Test get_standard_structure function with different set of configuration."""
    structure = get_standard_structure(element, prop, configuration)

    data_regression.check(structure.get_cif().get_content())



# -*- coding: utf-8 -*-
"""Equation of state workflow that can use any code plugin implementing the common relax workflow."""
from typing import List

from aiida import orm
from aiida.engine import WorkChain, append_, calcfunction, run_get_node
from aiida.plugins import CalculationFactory, WorkflowFactory

PwBaseWorkChain = WorkflowFactory("quantumespresso.pw.base")
birch_murnaghan_fit = CalculationFactory("sssp_workflow.birch_murnaghan_fit")


@calcfunction
def scale_structure(
    structure: orm.StructureData, scale_factor: orm.Float
) -> orm.StructureData:
    """Scale the structure with the given scaling factor."""
    ase = structure.get_ase().copy()
    ase.set_cell(ase.get_cell() * float(scale_factor) ** (1 / 3), scale_atoms=True)
    return orm.StructureData(ase=ase)


class _EquationOfStateWorkChain(WorkChain):
    """Workflow to compute the equation of state for a given crystal structure."""

    @classmethod
    def define(cls, spec):
        # yapf: disable
        super().define(spec)
        spec.input('structure', valid_type=orm.StructureData, help='The structure at equilibrium volume.')
        spec.input('kpoints_distance', valid_type=orm.Float, required=True,
            help='The kpoints distance used in generating the kmesh of unscaled structure then for all scaled structures')
        spec.input('scale_factors', valid_type=orm.List, required=False,
            help='The list of scale factors at which the volume and total energy of the structure should be computed.')
        spec.input('scale_count', valid_type=orm.Int, default=lambda: orm.Int(7),
            help='The number of points to compute for the equation of state.')
        spec.input('scale_increment', valid_type=orm.Float, default=lambda: orm.Float(0.02),
            help='The relative difference between consecutive scaling factors.')
        spec.expose_inputs(PwBaseWorkChain,
            exclude=('kpoints', 'pw.structure', 'pw.kpoints_distance'),
            namespace_options={'help': 'Inputs for the `PwBaseWorkChain` for the SCF calculation.'})
        spec.outline(
            cls.run_init,
            cls.run_eos,
            cls.inspect_eos,
        )
        spec.output('output_volume_energy', valid_type=orm.Dict,
            help='Results volumes and energise.')
        spec.output('output_birch_murnaghan_fit', valid_type=orm.Dict, required=False,
            help='Result of birch murnaghan fitting.')
        spec.exit_code(400, 'ERROR_SUB_PROCESS_FAILED',
            message='At least one of the `{cls}` sub processes did not finish successfully.')
        spec.exit_code(500, 'ERROR_BIRCH_MURNAGHAN_FIT_FAILED',
            message='The birch murnaghan fit failed with exit code={code}.')
        # yapf: enable

    def get_scale_factors(self) -> List[float]:
        """Return the list of scale factors.
        The points are averagely distributed from the minimal volume.
        The scale equal to 1 will not returned.

        Return: list
        """
        if "scale_factors" in self.inputs:
            return self.inputs.scale_factors.get_list()

        count = self.inputs.scale_count.value
        increment = self.inputs.scale_increment.value
        return [
            round(1 + (2 * i - count + 1) / 2 * increment, 3)
            for i in range(count)
            if (2 * i - count + 1 != 0)
        ]

    def _get_inputs(self, scale_factor):
        """Inputs for pw calculation of every scale increment."""
        inputs = self.exposed_inputs(PwBaseWorkChain)

        # structure scaled may lead to kpoints mesh change if it is generated from kpoint distance,
        # therefore the kpoints mesh is fixed for all scaled structures and from the unscaled one.
        kpoints = orm.KpointsData()
        kpoints.set_cell_from_structure(self.inputs.structure)
        kpoints.set_kpoints_mesh_from_density(
            distance=self.inputs.kpoints_distance.value
        )

        inputs["metadata"] = {"call_link_label": "EOS_scf"}
        inputs["kpoints"] = kpoints
        inputs["pw"]["structure"] = scale_structure(
            self.inputs.structure, orm.Float(scale_factor)
        )
        inputs.pop("kpoints_distance", None)

        return inputs

    def run_init(self):
        """Run the first sub-workchain, if this failed the whole workchain break."""
        inputs = self._get_inputs(scale_factor=1)  # inputs for unscaled structure
        self.report(f"submitting precheck calculation for unscaled structure.")
        self.ctx.init_workchain = self.submit(PwBaseWorkChain, **inputs)
        self.to_context(children=append_(self.ctx.init_workchain))

    def run_eos(self):
        """Run the sub process at each scale factor to compute the structure volume and total energy."""
        workchain = self.ctx.children[0]

        if not workchain.is_finished_ok:
            self.logger.warning(
                f"PwBaseWorkChain pk={workchain.pk} for first scale structure run is failed."
            )
            return self.exit_codes.ERROR_SUB_PROCESS_FAILED.format(cls=PwBaseWorkChain)

        for scale_factor in self.get_scale_factors():
            inputs = self._get_inputs(scale_factor)
            self.report(f"submitting scale_factor=`{scale_factor}`")
            self.to_context(children=append_(self.submit(PwBaseWorkChain, **inputs)))

    def inspect_eos(self):
        """Inspect all children workflows to make sure they finished successfully."""
        if any(not child.is_finished_ok for child in self.ctx.children):
            return self.exit_codes.ERROR_SUB_PROCESS_FAILED.format(cls=PwBaseWorkChain)

        volume_energy = {
            "num_of_atoms": sum(self.inputs.structure.get_composition().values()),
            "volume_unit": "A^3",
            "energy_unit": "eV",
        }
        volumes = []
        energies = []
        for child in self.ctx.children:
            # free energy E-TS (metal)
            energies.append(child.outputs.output_parameters["energy"])
            volumes.append(child.outputs.output_parameters["volume"])

        self.logger.info(f"volumes={volumes}, energies={energies}")

        volume_energy["volumes"] = volumes
        volume_energy["energies"] = energies

        output_volume_energy = orm.Dict(dict=volume_energy).store()
        self.out("output_volume_energy", output_volume_energy)

        output_birch_murnaghan_fit, node = run_get_node(
            birch_murnaghan_fit, output_volume_energy
        )

        if not node.is_finished_ok:
            self.logger.warning(f"The birch murnaghan fit failed for node pk={node.pk}")
            return self.exit_codes.ERROR_BIRCH_MURNAGHAN_FIT_FAILED.format(
                code=node.exit_status
            )
        else:
            self.report(
                f"The birch murnaghan fitting results are: {output_birch_murnaghan_fit.get_dict()}"
            )
            self.out("output_birch_murnaghan_fit", output_birch_murnaghan_fit)



# -*- coding: utf-8 -*-
"""
Birch-Murnaghan fit as calcfunction refactor from eosfit.py v3.1 write by Kurt Lejaeghere
Copyright (C) 2012 Kurt Lejaeghere <Kurt.Lejaeghere@UGent.be>, Center for
Molecular Modeling (CMM), Ghent University, Ghent, Belgium
"""
from aiida import orm
from aiida.engine import ExitCode, calcfunction


@calcfunction
def birch_murnaghan_fit(volume_energy: orm.Dict):
    """
    doc
    """
    # pylint: disable=invalid-name
    import numpy as np

    volumes = np.array(volume_energy["volumes"])
    energies = np.array(volume_energy["energies"])
    num_of_atoms = volume_energy["num_of_atoms"]
    fitdata = np.polyfit(volumes ** (-2.0 / 3.0), energies, 3, full=True)

    ssr = fitdata[1]
    sst = np.sum((energies - np.average(energies)) ** 2.0)
    residuals0 = ssr / sst
    deriv0 = np.poly1d(fitdata[0])
    deriv1 = np.polyder(deriv0, 1)
    deriv2 = np.polyder(deriv1, 1)
    deriv3 = np.polyder(deriv2, 1)

    volume0 = 0
    x = 0
    for x in np.roots(deriv1):
        if x > 0 and deriv2(x) > 0:
            volume0 = x ** (-3.0 / 2.0)
            break

    if volume0 == 0:
        return ExitCode(100, f"get spurious volume0={volume0}.")

    if not isinstance(volume0, float):
        # In case where the fitting failed
        return ExitCode(101, f"get spurious volume0={volume0}")

    derivV2 = 4.0 / 9.0 * x**5.0 * deriv2(x)
    derivV3 = -20.0 / 9.0 * x ** (13.0 / 2.0) * deriv2(x) - 8.0 / 27.0 * x ** (
        15.0 / 2.0
    ) * deriv3(x)
    bulk_modulus0 = derivV2 / x ** (3.0 / 2.0)
    bulk_deriv0 = -1 - x ** (-3.0 / 2.0) * derivV3 / derivV2
    energy0 = deriv0(volume0 ** (-2.0 / 3.0))

    echarge = 1.60217733e-19
    return orm.Dict(
        dict={
            "volume0": round(volume0, 7),
            "energy0": round(energy0, 7),
            "num_of_atoms": int(num_of_atoms),
            "bulk_modulus0": round(bulk_modulus0, 7),
            "bulk_modulus0_GPa": round(bulk_modulus0 * echarge * 1.0e21, 7),
            "bulk_deriv0": round(bulk_deriv0, 7),
            "residuals0": round(residuals0[0]),
            "volume0_unit": "A^3",
            "bulk_modulus0_unit": "eV A^3",
            "bulk_modulus0_GPa_unit": "GPa",
        }
    )



# -*- coding: utf-8 -*-
"""Workchain to compute a band structure for a given structure using Quantum ESPRESSO pw.x.
This is a modified version of the original `PwBandsWorkChain` in aiida-quantumespresso in order to
support caching of the results wisely. The PW calculation will be re-run if the the parent folder is cleaned and
but required by the following PH calculation.
"""
from aiida import orm
from aiida.common import AttributeDict
from aiida.engine import ToContext, WorkChain, if_, while_
from aiida.plugins import WorkflowFactory
from aiida_quantumespresso.calculations.functions.seekpath_structure_analysis import (
    seekpath_structure_analysis,
)
from aiida_quantumespresso.utils.mapping import prepare_process_inputs
from aiida_quantumespresso.workflows.protocols.utils import ProtocolMixin

PwBaseWorkChain = WorkflowFactory("quantumespresso.pw.base")
PwRelaxWorkChain = WorkflowFactory("quantumespresso.pw.relax")


def validate_inputs(inputs, ctx=None):  # pylint: disable=unused-argument
    """Validate the inputs of the entire input namespace."""
    # pylint: disable=no-member

    if "nbands_factor" in inputs and "nbnd" in inputs["bands"]["pw"][
        "parameters"
    ].base.attributes.get("SYSTEM", {}):
        return PwBandsWorkChain.exit_codes.ERROR_INVALID_INPUT_NUMBER_OF_BANDS.message

    # Cannot specify both `bands_kpoints` and `bands_kpoints_distance`
    if all(key in inputs for key in ["bands_kpoints", "bands_kpoints_distance"]):
        return PwBandsWorkChain.exit_codes.ERROR_INVALID_INPUT_KPOINTS.message


class PwBandsWorkChain(ProtocolMixin, WorkChain):
    """Workchain to compute a band structure for a given structure using Quantum ESPRESSO pw.x.

    The logic for the computation of various parameters for the BANDS step is as follows:

    Number of bands:
        One can specify the number of bands to be used in the BANDS step either directly through the input parameters
        `bands.pw.parameters.SYSTEM.nbnd` or through `nbands_factor`. Note that specifying both is not allowed. When
        neither is specified nothing will be set by the work chain and the default of Quantum ESPRESSO will end up being
        used. If the `nbands_factor` is specified the maximum value of the following values will be used:

        * `nbnd` of the preceding SCF calculation
        * 0.5 * nelectrons * nbands_factor
        * 0.5 * nelectrons + 4

    Kpoints:
        There are three options; specify either an existing `KpointsData` through `bands_kpoints`, or specify the
        `bands_kpoint_distance`, or specify neither. For the former those exact kpoints will be used for the BANDS step.
        In the two other cases, the structure will first be normalized using SeekPath and the path along high-symmetry
        k-points will be generated on that structure. The distance between kpoints for the path will be equal to that
        of `bands_kpoints_distance` or the SeekPath default if not specified.
    """

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        # yapf: disable
        super().define(spec)
        spec.expose_inputs(PwRelaxWorkChain, namespace='relax', exclude=('clean_workdir', 'structure'),
            namespace_options={'required': False, 'populate_defaults': False,
            'help': 'Inputs for the `PwRelaxWorkChain`, if not specified at all, the relaxation step is skipped.'})
        spec.expose_inputs(PwBaseWorkChain, namespace='scf',
            exclude=('clean_workdir', 'pw.structure'),
            namespace_options={'help': 'Inputs for the `PwBaseWorkChain` for the SCF calculation.'})
        spec.expose_inputs(PwBaseWorkChain, namespace='bands',
            exclude=('clean_workdir', 'pw.structure', 'pw.kpoints', 'pw.kpoints_distance', 'pw.parent_folder'),
            namespace_options={'help': 'Inputs for the `PwBaseWorkChain` for the BANDS calculation.'})
        spec.input('structure', valid_type=orm.StructureData, help='The inputs structure.')
        spec.input('clean_workdir', valid_type=orm.Bool, default=lambda: orm.Bool(False),
            help='If `True`, work directories of all called calculation will be cleaned at the end of execution.')
        spec.input('nbands_factor', valid_type=orm.Float, required=False,
            help='The number of bands for the BANDS calculation is that used for the SCF multiplied by this factor.')
        spec.input('bands_kpoints', valid_type=orm.KpointsData, required=False,
            help='Explicit kpoints to use for the BANDS calculation. Specify either this or `bands_kpoints_distance`.')
        spec.input('bands_kpoints_distance', valid_type=orm.Float, required=False,
            help='Minimum kpoints distance for the BANDS calculation. Specify either this or `bands_kpoints`.')
        spec.inputs.validator = validate_inputs
        spec.outline(
            cls.setup,
            if_(cls.should_run_relax)(
                cls.run_relax,
                cls.inspect_relax,
            ),
            if_(cls.should_run_seekpath)(
                cls.run_seekpath,
            ),
            while_(cls.should_run_band)(
                cls.run_scf,
                cls.inspect_scf,
                cls.run_bands,
                cls.inspect_bands,
            ),
            cls.results,
        )
        spec.exit_code(201, 'ERROR_INVALID_INPUT_NUMBER_OF_BANDS',
            message='Cannot specify both `nbands_factor` and `bands.pw.parameters.SYSTEM.nbnd`.')
        spec.exit_code(202, 'ERROR_INVALID_INPUT_KPOINTS',
            message='Cannot specify both `bands_kpoints` and `bands_kpoints_distance`.')
        spec.exit_code(401, 'ERROR_SUB_PROCESS_FAILED_RELAX',
            message='The PwRelaxWorkChain sub process failed')
        spec.exit_code(402, 'ERROR_SUB_PROCESS_FAILED_SCF',
            message='The scf PwBasexWorkChain sub process failed')
        spec.exit_code(403, 'ERROR_SUB_PROCESS_FAILED_BANDS',
            message='The bands PwBasexWorkChain sub process failed')
        spec.output('primitive_structure', valid_type=orm.StructureData,
            required=False,
            help='The normalized and primitivized structure for which the bands are computed.')
        spec.output('seekpath_parameters', valid_type=orm.Dict,
            required=False,
            help='The parameters used in the SeeKpath call to normalize the input or relaxed structure.')
        spec.output('scf_parameters', valid_type=orm.Dict,
            help='The output parameters of the SCF `PwBaseWorkChain`.')
        spec.output('band_parameters', valid_type=orm.Dict,
            help='The output parameters of the BANDS `PwBaseWorkChain`.')
        spec.output('band_structure', valid_type=orm.BandsData,
            help='The computed band structure.')
        # yapf: enable

    @classmethod
    def get_protocol_filepath(cls):
        """Return ``pathlib.Path`` to the ``.yaml`` file that defines the protocols."""
        from importlib_resources import files

        from ..protocols import pw as pw_protocols

        return files(pw_protocols) / "bands.yaml"

    @classmethod
    def get_builder_from_protocol(
        cls, code, structure, protocol=None, overrides=None, options=None, **kwargs
    ):
        """Return a builder prepopulated with inputs selected according to the chosen protocol.

        :param code: the ``Code`` instance configured for the ``quantumespresso.pw`` plugin.
        :param structure: the ``StructureData`` instance to use.
        :param protocol: protocol to use, if not specified, the default will be used.
        :param overrides: optional dictionary of inputs to override the defaults of the protocol.
        :param options: A dictionary of options that will be recursively set for the ``metadata.options`` input of all
            the ``CalcJobs`` that are nested in this work chain.
        :param kwargs: additional keyword arguments that will be passed to the ``get_builder_from_protocol`` of all the
            sub processes that are called by this workchain.
        :return: a process builder instance with all inputs defined ready for launch.
        """
        inputs = cls.get_protocol_inputs(protocol, overrides)

        args = (code, structure, protocol)
        relax = PwRelaxWorkChain.get_builder_from_protocol(
            *args, overrides=inputs.get("relax", None), options=options, **kwargs
        )
        scf = PwBaseWorkChain.get_builder_from_protocol(
            *args, overrides=inputs.get("scf", None), options=options, **kwargs
        )
        bands = PwBaseWorkChain.get_builder_from_protocol(
            *args, overrides=inputs.get("bands", None), options=options, **kwargs
        )

        relax.pop("structure", None)
        relax.pop("clean_workdir", None)
        relax.pop("base_final_scf", None)
        scf["pw"].pop("structure", None)
        scf.pop("clean_workdir", None)
        bands["pw"].pop("structure", None)
        bands.pop("clean_workdir", None)
        bands.pop("kpoints_distance", None)
        bands.pop("kpoints_force_parity", None)

        builder = cls.get_builder()
        builder.structure = structure
        builder.relax = relax
        builder.scf = scf
        builder.bands = bands
        builder.clean_workdir = orm.Bool(inputs["clean_workdir"])
        builder.nbands_factor = orm.Float(inputs["nbands_factor"])
        builder.bands_kpoints_distance = orm.Float(inputs["bands_kpoints_distance"])

        return builder

    def setup(self):
        """Define the current structure in the context to be the input structure."""
        self.ctx.current_structure = self.inputs.structure
        self.ctx.current_number_of_bands = None
        self.ctx.bands_kpoints = self.inputs.get("bands_kpoints", None)

        # Set to run for the first loop
        self.ctx.should_run_band = True

        # The list of nodes that are invalided the caching of the node and re-run scf calculation. When the remote folder of parent calculation is empty, the caching is invalid, but since invalid caching is applied to all_same_nodes, we need to keep track of the invalid nodes and bring caching valid back to the node so the next run will still using the caching, otherwise the next parent calculation will be re-run again, which is not what we want.
        # Warning!! experimental feature and will potentially cause "race condition" like behavior between workflows.
        self.ctx.cache_invalid_list = []

    def should_run_band(self):
        """The value will be set to `True` if bands calculation is failed because of the remote folder issue.
        Otherwise it keeps on trying to rerun the bands calculation until remote folder of preliminary scf is not found.
        """
        return self.ctx.should_run_band

    def should_run_relax(self):
        """If the 'relax' input namespace was specified, we relax the input structure."""
        return "relax" in self.inputs

    def should_run_seekpath(self):
        """Seekpath should only be run if the `bands_kpoints` input is not specified."""
        return "bands_kpoints" not in self.inputs

    def run_relax(self):
        """Run the PwRelaxWorkChain to run a relax PwCalculation."""
        inputs = AttributeDict(self.exposed_inputs(PwRelaxWorkChain, namespace="relax"))
        inputs.metadata.call_link_label = "relax"
        inputs.structure = self.ctx.current_structure

        running = self.submit(PwRelaxWorkChain, **inputs)

        self.report(f"launching PwRelaxWorkChain<{running.pk}>")

        return ToContext(workchain_relax=running)

    def inspect_relax(self):
        """Verify that the PwRelaxWorkChain finished successfully."""
        workchain = self.ctx.workchain_relax

        if not workchain.is_finished_ok:
            self.report(
                f"PwRelaxWorkChain failed with exit status {workchain.exit_status}"
            )
            return self.exit_codes.ERROR_SUB_PROCESS_FAILED_RELAX

        self.ctx.current_structure = workchain.outputs.output_structure
        self.ctx.current_number_of_bands = (
            workchain.outputs.output_parameters.base.attributes.get("number_of_bands")
        )

    def run_seekpath(self):
        """Run the structure through SeeKpath to get the normalized structure and path along high-symmetry k-points .

        This is only called if the `bands_kpoints` input was not specified.
        """
        inputs = {
            "reference_distance": self.inputs.get("bands_kpoints_distance", None),
            "metadata": {"call_link_label": "seekpath"},
        }
        result = seekpath_structure_analysis(self.ctx.current_structure, **inputs)
        self.ctx.current_structure = result["primitive_structure"]
        self.ctx.bands_kpoints = result["explicit_kpoints"]

        self.out("primitive_structure", result["primitive_structure"])
        self.out("seekpath_parameters", result["parameters"])

    def run_scf(self):
        """Run the PwBaseWorkChain in scf mode on the primitive cell of (optionally relaxed) input structure."""
        inputs = AttributeDict(self.exposed_inputs(PwBaseWorkChain, namespace="scf"))
        inputs.metadata.call_link_label = "scf"
        inputs.pw.structure = self.ctx.current_structure
        inputs.pw.parameters = inputs.pw.parameters.get_dict()
        inputs.pw.parameters.setdefault("CONTROL", {})["calculation"] = "scf"

        # Make sure to carry the number of bands from the relax workchain if it was run and it wasn't explicitly defined
        # in the inputs. One of the base workchains in the relax workchain may have changed the number automatically in
        #  the sanity checks on band occupations.
        if self.ctx.current_number_of_bands:
            inputs.pw.parameters.setdefault("SYSTEM", {}).setdefault(
                "nbnd", self.ctx.current_number_of_bands
            )

        inputs = prepare_process_inputs(PwBaseWorkChain, inputs)
        running = self.submit(PwBaseWorkChain, **inputs)

        self.report(f"launching PwBaseWorkChain<{running.pk}> in scf mode")

        return ToContext(workchain_scf=running)

    def inspect_scf(self):
        """Verify that the PwBaseWorkChain for the scf run finished successfully."""
        workchain = self.ctx.workchain_scf

        # duplicate of _disable_cache of _BaseEvaluateWorkChain
        # I do not want SCF calc used for caching if it is a from cached calculation
        # right after!@@!
        if workchain.is_finished:
            for child in workchain.called_descendants:
                if (
                    child.process_label == "PwCalculation"
                    and child.base.caching.is_created_from_cache
                ):
                    child.is_valid_cache = False

        if not workchain.is_finished_ok:
            self.report(
                f"scf PwBaseWorkChain failed with exit status {workchain.exit_status}"
            )
            return self.exit_codes.ERROR_SUB_PROCESS_FAILED_SCF

        self.ctx.current_folder = workchain.outputs.remote_folder
        # the current_folder is set so can bring the caching invalid back to the node
        while self.ctx.cache_invalid_list:
            node_pk = self.ctx.cache_invalid_list.pop()
            node = orm.load_node(node_pk)
            node.is_valid_cache = True

        self.ctx.current_number_of_bands = (
            workchain.outputs.output_parameters.base.attributes.get("number_of_bands")
        )

    def run_bands(self):
        """Run the PwBaseWorkChain in bands mode along the path of high-symmetry determined by seekpath."""
        inputs = AttributeDict(self.exposed_inputs(PwBaseWorkChain, namespace="bands"))
        inputs.metadata.call_link_label = "bands"
        inputs.kpoints = self.ctx.bands_kpoints
        inputs.pw.structure = self.ctx.current_structure
        inputs.pw.parent_folder = self.ctx.current_folder
        inputs.pw.parameters = inputs.pw.parameters.get_dict()
        inputs.pw.parameters.setdefault("CONTROL", {})
        inputs.pw.parameters.setdefault("SYSTEM", {})
        inputs.pw.parameters.setdefault("ELECTRONS", {})

        # The following flags always have to be set in the parameters, regardless of what caller specified in the inputs
        inputs.pw.parameters["CONTROL"]["calculation"] = "bands"

        # Only set the following parameters if not directly explicitly defined in the inputs
        inputs.pw.parameters["ELECTRONS"].setdefault("diagonalization", "cg")
        inputs.pw.parameters["ELECTRONS"].setdefault("diago_full_acc", True)

        # If `nbands_factor` is defined in the inputs we set the `nbnd` parameter
        if "nbands_factor" in self.inputs:
            factor = self.inputs.nbands_factor.value
            parameters = self.ctx.workchain_scf.outputs.output_parameters.get_dict()
            nbands = int(parameters["number_of_bands"])
            nelectron = int(parameters["number_of_electrons"])
            nbnd = max(int(0.5 * nelectron * factor), int(0.5 * nelectron) + 4, nbands)
            inputs.pw.parameters["SYSTEM"]["nbnd"] = nbnd

        # Otherwise set the current number of bands, unless explicitly set in the inputs
        else:
            inputs.pw.parameters["SYSTEM"].setdefault(
                "nbnd", self.ctx.current_number_of_bands
            )

        inputs = prepare_process_inputs(PwBaseWorkChain, inputs)
        running = self.submit(PwBaseWorkChain, **inputs)

        self.report(f"launching PwBaseWorkChain<{running.pk}> in bands mode")

        return ToContext(workchain_bands=running)

    def inspect_bands(self):
        """Verify that the PwBaseWorkChain for the bands run finished successfully."""
        workchain = self.ctx.workchain_bands

        if not workchain.is_finished_ok:
            # The inner logic to detect if the failed bands calculation is due to the empty remote folder
            # If so, we will invalid the caching of the node and re-run scf calculation.
            # !!! This empty check is not a good way to detect the remote folder issue, but it is the only way I can think of now.
            # !!! If the SSH connection is lost, the exception will be raised and the workchain will be terminated.
            # Therefore use cleaned flag to detect the remote folder issue this didn't require SSH connection. But will not work if the remote folder is cleaned by other means.
            is_cleaned = self.ctx.current_folder.base.extras.get("cleaned", False)
            if is_cleaned:
                self.logger.warning(
                    f"PhBaseWorkChain failed because the remote folder is empty with exit status {workchain.exit_status}, invalid the caching of the node and re-run scf calculation."
                )
                # invalid the caching of the node and re-run scf calculation
                workchain_scf = self.ctx.workchain_scf
                scf_pw_node = [
                    c for c in workchain_scf.called if isinstance(c, orm.CalcJobNode)
                ][0]
                all_same_nodes = scf_pw_node.base.caching.get_all_same_nodes()
                for node in all_same_nodes:
                    node.is_valid_cache = False
                    self.ctx.cache_invalid_list.append(node.pk)
                return
            else:
                self.report(
                    f"bands PwBaseWorkChain failed with exit status {workchain.exit_status}"
                )
                return self.exit_codes.ERROR_SUB_PROCESS_FAILED_BANDS

        # The bands workchain finished successfully, so we can set the `should_run_band` flag to False
        self.ctx.should_run_band = False

    def results(self):
        """Attach the desired output nodes directly as outputs of the workchain."""
        self.report("workchain succesfully completed")
        self.out("scf_parameters", self.ctx.workchain_scf.outputs.output_parameters)
        self.out("band_parameters", self.ctx.workchain_bands.outputs.output_parameters)
        self.out("band_structure", self.ctx.workchain_bands.outputs.output_band)

    def on_terminated(self):
        """Clean the working directories of all child calculations if `clean_workdir=True` in the inputs."""
        super().on_terminated()

        if self.inputs.clean_workdir.value is False:
            self.report("remote folders will not be cleaned")
            return

        cleaned_calcs = []

        for called_descendant in self.node.called_descendants:
            if isinstance(called_descendant, orm.CalcJobNode):
                try:
                    called_descendant.outputs.remote_folder._clean()  # pylint: disable=protected-access
                    cleaned_calcs.append(called_descendant.pk)
                except (IOError, OSError, KeyError):
                    pass

        if cleaned_calcs:
            self.report(
                f"cleaned remote folders of calculations: {' '.join(map(str, cleaned_calcs))}"
            )



# -*- coding: utf-8 -*-
"""
WorkChain calculate phonon frequencies at Gamma
"""
from aiida import orm
from aiida.common import NotExistentAttributeError
from aiida.engine import ToContext, while_
from aiida.plugins import DataFactory, WorkflowFactory

from aiida_sssp_workflow.workflows.common import clean_workdir, operate_calcjobs

from . import _BaseEvaluateWorkChain

PwBaseWorkflow = WorkflowFactory("quantumespresso.pw.base")
PhBaseWorkflow = WorkflowFactory("quantumespresso.ph.base")
UpfData = DataFactory("pseudo.upf")


class PhononFrequenciesWorkChain(_BaseEvaluateWorkChain):
    """WorkChain to calculate cohisive energy of input structure"""

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        # yapf: disable
        super().define(spec)
        spec.expose_inputs(PwBaseWorkflow, namespace='scf', include=['metadata', 'pw', 'kpoints_distance'])
        spec.expose_inputs(PhBaseWorkflow, namespace='phonon', exclude=['ph.parent_folder'])

        spec.outline(
            cls.setup,
            while_(cls.should_run_ph)(
                cls.run_scf,
                cls.inspect_scf,
                cls.run_ph,
                cls.inspect_ph,
            ),
            cls.finalize,
        )
        spec.output('output_parameters', valid_type=orm.Dict, required=True,
                    help='The output parameters include phonon frequencies.')
        spec.exit_code(211, 'ERROR_NO_REMOTE_FOLDER_OUTPUT_OF_SCF',
                    message='The remote folder node not exist')
        spec.exit_code(202, 'ERROR_SUB_PROCESS_FAILED_PH',
                    message='The `PhBaseWorkChain` sub process failed.')
        spec.exit_code(203, 'ERROR_SUB_PROCESS_FAILED_SCF',
                    message='The `PwBaseWorkChain` sub process failed.')
        # yapf: enable

    def setup(self):
        self.ctx.should_run_ph = True

        # The list of nodes that are invalided the caching of the node and re-run scf calculation. When the remote folder of parent calculation is empty, the caching is invalid, but since invalid caching is applied to all_same_nodes, we need to keep track of the invalid nodes and bring caching valid back to the node so the next run will still using the caching, otherwise the next parent calculation will be re-run again, which is not what we want.
        # Warning!! experimental feature and will potentially cause "race condition" like behavior between workflows.
        self.ctx.cache_invalid_list = []

    def should_run_ph(self):
        """will be True if ph calculation is successful, otherwise will rerun scf
        The ph calculation can be failed due to the following reasons:
        - the scf calculation is not successful
        - the ph calculation is not successful
        - the scf calculation is successful but the remote folder is not found and the ph can not be submitted
        """
        return self.ctx.should_run_ph

    def run_scf(self):
        """
        set the inputs and submit scf
        """
        inputs = self.exposed_inputs(PwBaseWorkflow, namespace="scf")

        running = self.submit(PwBaseWorkflow, **inputs)
        self.report(f"Running pw calculation pk={running.pk}")
        return ToContext(workchain_scf=running)

    def inspect_scf(self):
        """inspect the result of scf calculation if fail do not continue."""
        workchain = self.ctx.workchain_scf

        if workchain.is_finished:
            self._disable_cache(workchain)

        if not workchain.is_finished_ok:
            self.logger.warning(
                f"PwBaseWorkChain failed with exit status {workchain.exit_status}"
            )

            return self.exit_codes.ERROR_SUB_PROCESS_FAILED_SCF

        try:
            remote_folder = self.ctx.scf_remote_folder = workchain.outputs.remote_folder

            self.report(
                f"Is remote folder cleaned? {remote_folder.extras.get('cleaned', False)}"
            )
        except NotExistentAttributeError:
            # set condition to False to break loop
            return self.exit_codes.ERROR_NO_REMOTE_FOLDER_OUTPUT_OF_SCF

        # When change caching below, also check _bands and _caching_wise_bandes
        # the remote folder is set so can bring the caching invalid back to the node
        while self.ctx.cache_invalid_list:
            node_pk = self.ctx.cache_invalid_list.pop()
            node = orm.load_node(node_pk)
            node.is_valid_cache = True

        self.ctx.ecutwfc = workchain.inputs.pw.parameters["SYSTEM"]["ecutwfc"]
        self.ctx.ecutrho = workchain.inputs.pw.parameters["SYSTEM"]["ecutrho"]

        self.ctx.calc_time = workchain.outputs.output_parameters["wall_time_seconds"]

    def run_ph(self):
        """
        set the inputs and submit ph calculation to get quantities for phonon evaluation
        """

        inputs = self.exposed_inputs(PhBaseWorkflow, namespace="phonon")
        inputs["ph"]["parent_folder"] = self.ctx.scf_remote_folder

        running = self.submit(PhBaseWorkflow, **inputs)
        self.report(f"Running ph calculation pk={running.pk}")
        return ToContext(workchain_ph=running)

    def inspect_ph(self):
        """inspect the result of ph calculation."""
        workchain = self.ctx.workchain_ph

        if not workchain.is_finished_ok:
            # if the remote folder is empty, invalid the caching of the node and re-run scf calculation
            is_cleaned = self.ctx.scf_remote_folder.base.extras.get("cleaned", False)
            if is_cleaned:
                self.logger.warning(
                    f"PhBaseWorkChain failed because the remote folder is empty with exit status {workchain.exit_status}, invalid the caching of the node and re-run scf calculation."
                )
                # invalid the caching of the node and re-run scf calculation
                workchain_scf = self.ctx.workchain_scf
                pw_node = [
                    c for c in workchain_scf.called if isinstance(c, orm.CalcJobNode)
                ][0]
                all_same_nodes = pw_node.base.caching.get_all_same_nodes()
                for node in all_same_nodes:
                    node.is_valid_cache = False
                    self.ctx.cache_invalid_list.append(node.pk)
                return
            else:
                self.report(
                    f"PhBaseWorkChain for pressure evaluation failed with exit status {workchain.exit_status}"
                )
                return self.exit_codes.ERROR_SUB_PROCESS_FAILED_PH

        self.ctx.calc_time += workchain.outputs.output_parameters["wall_time_seconds"]
        output_parameters = workchain.outputs.output_parameters.get_dict()
        output_parameters.update(
            {
                "total_calc_time": self.ctx.calc_time,
                "time_unit": "s",
            }
        )
        self.out("output_parameters", orm.Dict(dict=output_parameters).store())

        self.ctx.should_run_ph = False

    def finalize(self):
        """set ecutwfc and ecutrho"""
        self.out("ecutwfc", orm.Int(self.ctx.ecutwfc).store())
        self.out("ecutrho", orm.Int(self.ctx.ecutrho).store())

        if self.inputs.clean_workdir.value is True:
            cleaned_calcs = operate_calcjobs(
                self.node, operator=clean_workdir, all_same_nodes=False
            )

            if cleaned_calcs:
                self.report(
                    f"cleaned remote folders of calculations: {' '.join(map(str, cleaned_calcs))}"
                )

        else:
            self.report(f"{type(self)}: remote folders will not be cleaned")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Extract SSSP verification results from AiiDA database.

This script will get verification nodes pk as inputs and export all results into a json file
It is only a template solution for how to store the results for showing in
the aiidalab-sssp app. In the long term, I plan to transfor these json file into mongodb
database and held a API on the dokku server able to be queried for detailed specific field of every pseudopotentials.
For the large data such as bandstructure, they are store on the materials cloud and
for aiidalab-sssp app, the results will be queried and retrieved from there by using restapi.

The json exported from this script will contain the uuid of the verification, and the outputs
from all its output namespaces.

One requirement of this script is when running on new verification node of the same pseudopotential, the outputs need to update with the sub-workflows with latest run but not the whole verification output.

The bands structure of bands data is extracted to a folder the json files with label as basename. The format should be that can directly used for bandsplot.

The bands data and bands parameters for bands distance need to store as one json for distance calculation.

If output json file exist, will abort only if --update flag is set. It will check the ctime and hash and update to the latest results.
Therefore, the logic to run the new verficiation is that the properties only run for necessary workflows but not the already good one.
Since the latest workflow results will always update the old one.
But it is okay and I can still do that.
"""

import json
import os
import tarfile
from pathlib import Path

import aiida
import click
from aiida import orm
from aiida.common import AttributeDict
from monty.json import jsanitize
from tqdm import tqdm

from aiida_sssp_workflow.cli import cmd_root

process_prop_label_mapping = {
    "cohesive_energy": "ConvergenceCohesiveEnergyWorkChain",
    "phonon_frequencies": "ConvergencePhononFrequenciesWorkChain",
    "pressure": "ConvergencePressureWorkChain",
    "bands": "ConvergenceBandsWorkChain",
    "delta": "ConvergenceDeltaWorkChain",
}


def export_bands_structure(band_structure, band_parameters):
    data = json.loads(band_structure._exportcontent("json", comments=False)[0])
    data["fermi_level"] = band_parameters["fermi_energy"]
    data["number_of_electrons"] = band_parameters["number_of_electrons"]
    data["number_of_bands"] = band_parameters["number_of_bands"]

    return jsanitize(data)


def export_bands_data(band_structure: orm.BandsData, band_parameters: orm.Dict):
    bands_arr = band_structure.get_bands()
    kpoints_arr, weights_arr = band_structure.get_kpoints(also_weights=True)

    data = {
        "fermi_level": band_parameters["fermi_energy"],
        "number_of_electrons": band_parameters["number_of_electrons"],
        "number_of_bands": band_parameters["number_of_bands"],
        "bands": bands_arr.tolist(),
        "kpoints": kpoints_arr.tolist(),  # TODO: using override JSON encoder
        "weights": weights_arr.tolist(),
    }

    return data


def _flatten_output(attr_dict, skip: list = []):
    """
    flaten output dict node
    node_collection is a list to accumulate the nodes that not unfolded

    :param skip: is a list of keys (format with parent_key.key) of Dict name that
        will not collected into the json file.

    For output nodes not being expanded, write down the uuid and datatype for future query.
    """
    # do_not_unfold = ["band_parameters", "scf_parameters", "seekpath_parameters"]

    for key, value in attr_dict.items():
        if key in skip:
            continue

        if isinstance(value, AttributeDict):
            # keep on unfold if it is a namespace
            _flatten_output(value, skip)
        elif isinstance(value, orm.Dict):
            attr_dict[key] = value.get_dict()
        elif isinstance(value, orm.Int):
            attr_dict[key] = value.value
        else:
            # node type not handled attach uuid
            attr_dict[key] = {
                "uuid": value.uuid,
                "datatype": type(value),
            }

    # print(archive_uuids)
    return attr_dict


def get_metadata(node):
    return {
        "uuid": node.uuid,
        "ctime": node.ctime,
        "_aiida_hash": node.get_hash(),
    }


@cmd_root.command("extract")
@click.option("element", "-e", help="element")
@click.option(
    "override",
    "--override",
    is_flag=True,
    default=False,
    help="Purge previous results of element, careful to use.",
)
@click.option("--dst", type=click.Path(), help="folder to store the output.")
@click.option(
    "custom_label",
    "--custom-label",
    required=False,
    help="custumrized label as key for pseudos the output json file.",
)
@click.argument("pks", nargs=-1)
def extract(pks, element, dst, override, custom_label):
    """Extract the results from the verification nodes and export to json file."""
    _profile = aiida.load_profile()
    click.echo(f"Profile: {_profile.name}")

    # can't use custom_label with more than one node output
    if custom_label and len(pks) > 1:
        raise ValueError("cant use label with more than one node.")

    if not dst:
        dst = "./sssp_db"

    # if no element provide, element==None, only pack the db to tar file
    if element:
        json_fn = f"{element}.json"
        Path(os.path.join(dst, "bands", element)).mkdir(parents=True, exist_ok=True)
        Path(os.path.join(dst, "band_structure", element)).mkdir(
            parents=True, exist_ok=True
        )
        Path(os.path.join(dst, json_fn)).touch(exist_ok=True)

        # read the results from the <element>.json, if not exist, create the json and set to {}
        try:
            with open(os.path.join(dst, json_fn), "r") as fh:
                curated_result = json.load(fh)
        except json.decoder.JSONDecodeError:
            # newly create empty file
            curated_result = {}

        for pk in tqdm(pks):
            # meta info of such verification
            # uuid, label
            _node = orm.load_node(pk)
            assert _node.get_attribute("process_label") == "VerificationWorkChain"
            label = (
                custom_label or _node.extras.get("label").split()[-1]
            )  # do not contain the extra machine info
            assert element == label.split(".")[0]

            # TODO label check

            # if override or newly run element
            if override or label not in curated_result:
                psp_result = {
                    "_metadata": [get_metadata(_node)],
                }  # the results of one verification
                psp_result["measure"] = {}
                psp_result["convergence"] = {}
            else:
                # append the new _metadata of the output node if it update the previous result
                psp_result = curated_result[label]
                metadata = get_metadata(_node)
                if metadata["_aiida_hash"] in [
                    i.get("_aiida_hash") for i in psp_result["_metadata"]
                ]:
                    click.echo(f"_metadata of {label} is in the list, do nothing.")
                    continue
                else:
                    psp_result["_metadata"] += [get_metadata(_node)]

            for called_wf in _node.called:
                if called_wf.process_label == "parse_pseudo_info":
                    psp_result["pseudo_info"] = {
                        **called_wf.outputs.result.get_dict(),
                    }
                # precision (EOS)
                if called_wf.process_label == "PrecisionMeasureWorkChain":
                    psp_result["measure"]["precision"] = _flatten_output(
                        _node.outputs.measure.precision
                    )
                    psp_result["measure"]["precision"]["_metadata"] = get_metadata(
                        called_wf
                    )
                # bands
                if called_wf.process_label == "BandsMeasureWorkChain":
                    psp_result["measure"]["bands"] = {
                        "bands": f"bands/{element}/{label}.json",
                        "band_structure": f"band_structure/{element}/{label}.json",
                    }
                    psp_result["measure"]["bands"]["_metadata"] = get_metadata(
                        called_wf
                    )

                    try:
                        with open(
                            os.path.join(dst, "bands", element, f"{label}.json"), "w"
                        ) as fh:
                            bands = called_wf.outputs.bands
                            json.dump(
                                export_bands_data(
                                    bands.band_structure, bands.band_parameters
                                ),
                                fh,
                            )
                        with open(
                            os.path.join(
                                dst, "band_structure", element, f"{label}.json"
                            ),
                            "w",
                        ) as fh:
                            bands = called_wf.outputs.band_structure
                            json.dump(
                                export_bands_structure(
                                    bands.band_structure, bands.band_parameters
                                ),
                                fh,
                            )
                    except Exception:
                        # bands not finished okay
                        pass

                # convergence
                for k, v in process_prop_label_mapping.items():
                    if called_wf.process_label == v:
                        try:
                            output_res = _flatten_output(_node.outputs.convergence[k])
                        except KeyError:
                            # run but not finished therefore no output node
                            output_res = {"message": "error"}
                        psp_result["convergence"][k] = output_res
                        psp_result["convergence"][k]["_metadata"] = get_metadata(
                            called_wf
                        )

            curated_result[f"{label}"] = psp_result

        with open(os.path.join(dst, json_fn), "w") as fh:
            json.dump(dict(curated_result), fh, indent=2, sort_keys=True, default=str)

    else:
        # if no element and pks as argument, simply pach for new version
        # # resursive to imply same order and use bz2 since no timestamp that
        # generate unidentical tarball.
        click.echo("Compressing...")
        with tarfile.open(f"{dst}.tar.gz", "w:bz2") as tar:
            tar.add(dst, arcname=os.path.basename(dst), recursive=True)


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
"""
Adapt from aiida-sssp-workflow as prototype
Will in here refactoring it so no need to use the aiida datatype as inputs.

calculate bands distance
"""
import numpy as np

# from aiida_sssp_workflow.efermi import find_efermi


def get_homo(bands, num_electrons: int):
    """
    This function only work for insulator,
    therefore the num_electrons is even number.
    """
    assert num_electrons % 2 == 0, f"There are {num_electrons} electrons, must be metal"
    # get homo band
    band = bands[:, num_electrons // 2 - 1]
    return max(band)


def fermi_dirac(band_energy, fermi_energy, smearing, spin):
    """
    The first argument can be an array
    """
    if spin:
        occ = 1.0
    else:
        occ = 2.0

    old_settings = np.seterr(over="raise", divide="raise")
    try:
        res = occ / (np.exp((band_energy - fermi_energy) / smearing) + 1.0)
    except FloatingPointError:
        res = np.heaviside(fermi_energy - band_energy, occ)
    np.seterr(**old_settings)

    return res


def retrieve_bands(
    bandsdata: dict,  # bands, kpoints, weights -> corresponding array
    start_band_idx,
    num_bands,
    num_electrons,
    smearing,
    do_smearing,
):
    """
    collect the bands of certain number with setting the start band.
    In order to make sure that when comparing two bands distance the number of bands is the same.

    aligh bands to fermi level

    The bands calculation of magnetic elements will giving a three dimensional bands where the
    first dimension is for the up and down spin.
    I simply concatenate along the first dimension.
    """
    bands = bandsdata.get("bands")
    # weights = bandsdata.get("weights")

    # reduce by first dimension of up, down spins
    if len(bands.shape) > 2:
        bands = bands[:, :, start_band_idx : start_band_idx + num_bands]
        nspin, nk, nbands = bands.shape
        bands = bands.reshape(nk, nbands * nspin)
    else:
        # update bands shift to fermi_level
        bands = bands[:, start_band_idx : start_band_idx + num_bands]

    # shift to fermi level aligh to zero
    bandsdata["bands"] = bands - bandsdata["fermi_level"]
    bandsdata["fermi_level"] = 0.0

    # update fermi_level
    if not do_smearing:
        # easy to spot the efermi energy only used for non-metals of typical configurations
        # in bands measure.
        homo_energy = get_homo(bands, num_electrons)
        bandsdata["fermi_level"] = homo_energy

    else:
        # for bands distance convergence
        # and metals in bands measure verification.
        # bands = np.asfortranarray(bands)
        # meth = 2  # firmi-dirac smearing

        # bandsdata["fermi_level"] = find_efermi(
        #     bands, weights, num_electrons, smearing, meth
        # )
        #####
        # use the fermi_level from QE therefore do nothing.
        # This can be commented out since with acwf protocol I use the fermi dirac smearing
        # which should give the exact the same fermi level.
        # We align the bands to fermi level and make the fermi level equal to 0.0 Ry as
        # code above.
        pass

    return bandsdata


def calculate_eta_and_max_diff(
    bandsdata_a: dict,
    bandsdata_b: dict,
    spin: bool,
    fermi_shift,
    smearing,
):
    """
    calculate the difference of two bands, weight is supported
    """
    from functools import partial

    from scipy.optimize import minimize

    weight_a = bandsdata_a.get("weights")
    weight_b = bandsdata_b.get("weights")
    weight = weight_a
    assert np.allclose(
        weight_a, weight_b
    ), "Different weight of kpoints of two calculation."

    bands_a = bandsdata_a.get("bands")
    bands_b = bandsdata_b.get("bands")

    num_bands = min(np.shape(bands_a)[1], np.shape(bands_b)[1])

    assert np.shape(bands_a)[0] == np.shape(bands_b)[0], "have different kpoints"

    # truncate the bands to same size
    bands_a = bands_a[:, : num_bands - 1]
    bands_b = bands_b[:, : num_bands - 1]

    # all bands are already shifted to fermi level aligh to zero
    occ_a = fermi_dirac(bands_a, fermi_shift, smearing, spin)
    occ_b = fermi_dirac(bands_b, fermi_shift, smearing, spin)
    occ = np.sqrt(occ_a * occ_b)

    bands_diff = bands_a - bands_b

    def fun_shift(occ, bands_diff, shift):
        # 1/w ~ degeneracy of the kpoints
        nominator = np.multiply(weight[:, None], (occ * (bands_diff + shift) ** 2))
        denominator = np.multiply(weight[:, None], occ)
        return np.sqrt(np.sum(nominator) / np.sum(denominator))

    # Compute eta
    eta_val = partial(fun_shift, occ, bands_diff)
    results = minimize(eta_val, np.array([0.0]), method="Nelder-Mead")
    eta = results.get("fun")
    shift = results.get("x")[0]

    # Compute max_diff:
    # then find from abs_diff the max one of which the occ > 0.5
    # first make occ > 0.5 to be 1 and occ <= 0.5 to be 0, then element-wise the
    # new occ matrix with abs_diff, and find the max diff
    abs_diff = np.abs(bands_diff + shift)
    new_occ = np.heaviside(occ - 0.5, 0.0)
    new_diff = np.multiply(abs_diff, new_occ)
    max_diff = np.amax(new_diff)

    return {
        "eta": eta,
        "shift": shift,
        "max_diff": max_diff,
    }


def get_bands_distance(
    bandsdata_a: dict,
    bandsdata_b: dict,
    smearing: float,  # from degauss
    fermi_shift: float,
    do_smearing: bool,
    spin: bool,
):
    """
    example of bandsdata_a -> dict = {
        "number_of_electrons": 10,
        "number_of_bands": 10,
        "fermi_level": -0.97,
        "bands": <bands array> as list,
        "kpoints": <kpoints array> as list,
        "weights": <weights array> as list,
    }


    First aligh the number of two bands, e.g tranctrate the overceed nubmer of bands
    """
    _RY_TO_EV = 13.6056980659
    smearing = smearing * _RY_TO_EV

    # post process to deserial list to numpy arrar
    for key in ["bands", "kpoints", "weights"]:
        bandsdata_a[key] = np.asarray(bandsdata_a[key])
        bandsdata_b[key] = np.asarray(bandsdata_b[key])

    # make sure always less electrons bands as a. b hase more electrons if not equal
    if not int(bandsdata_b["number_of_electrons"]) >= int(
        bandsdata_a["number_of_electrons"]
    ):
        # swap to make sure a is less electrons pseudo
        bandsdata_a, bandsdata_b = bandsdata_b, bandsdata_a

    assert int(bandsdata_b["number_of_electrons"]) >= int(
        bandsdata_a["number_of_electrons"]
    ), f"Need to be less num_bands in a {bandsdata_a['number_of_electrons']} than b {bandsdata_b['number_of_electrons']}"

    num_electrons_a = int(bandsdata_a["number_of_electrons"])
    num_electrons_b = int(bandsdata_b["number_of_electrons"])

    # divide by 2 is valid for both spin and non-spin bands, since for spin I concatenate the bands
    # the number of bands is half of electrons
    band_b_start_band = int(num_electrons_b - num_electrons_a) // 2

    num_bands_a = bandsdata_a["number_of_bands"]
    num_bands_b = bandsdata_b["number_of_bands"] - band_b_start_band

    num_bands = min(num_bands_a, num_bands_b)
    num_electrons = min(num_electrons_a, num_electrons_b)

    bandsdata_a = retrieve_bands(
        bandsdata_a, 0, num_bands, num_electrons, smearing, do_smearing
    )

    bandsdata_b = retrieve_bands(
        bandsdata_b,
        band_b_start_band,
        num_bands,
        num_electrons,
        smearing,
        do_smearing,
    )

    # after cut and aligh in retrive band, the shapes are same now
    assert np.shape(bandsdata_a["bands"]) == np.shape(
        bandsdata_b["bands"]
    ), f'{np.shape(bandsdata_a["bands"])} != {np.shape(bandsdata_b["bands"])}'

    # eta_v
    fermi_shift_v = 0.0
    if do_smearing:
        smearing_v = smearing
    else:
        smearing_v = 0

    outputs = calculate_eta_and_max_diff(
        bandsdata_a, bandsdata_b, spin, fermi_shift_v, smearing_v
    )

    _eV_to_mev = 1000
    eta_v = outputs.get("eta") * _eV_to_mev
    shift_v = outputs.get("shift") * _eV_to_mev
    max_diff_v = outputs.get("max_diff") * _eV_to_mev

    # eta_c
    # if not metal
    smearing_c = smearing
    outputs = calculate_eta_and_max_diff(
        bandsdata_a, bandsdata_b, spin, fermi_shift, smearing_c
    )

    eta_c = outputs.get("eta") * _eV_to_mev
    shift_c = outputs.get("shift") * _eV_to_mev
    max_diff_c = outputs.get("max_diff") * _eV_to_mev

    out = {
        "eta_v": eta_v,
        "shift_v": shift_v,
        "max_diff_v": max_diff_v,
        "eta_c": eta_c,
        "shift_c": shift_c,
        "max_diff_c": max_diff_c,
        "units": "meV",
    }

    return out
